"""Nigerian constants: states, LGAs, banks, telcos, and regex patterns.

Provides reference data and utilities for Nigerian-specific validation and formatting.
"""
from __future__ import annotations

import re
from typing import Dict, List, Optional

# =============================================================================
# Nigerian States and Capitals (36 states + FCT)
# =============================================================================

STATES: Dict[str, str] = {
    "Abia": "Umuahia",
    "Adamawa": "Yola",
    "Akwa Ibom": "Uyo",
    "Anambra": "Awka",
    "Bauchi": "Bauchi",
    "Bayelsa": "Yenagoa",
    "Benue": "Makurdi",
    "Borno": "Maiduguri",
    "Cross River": "Calabar",
    "Delta": "Asaba",
    "Ebonyi": "Abakaliki",
    "Edo": "Benin City",
    "Ekiti": "Ado Ekiti",
    "Enugu": "Enugu",
    "FCT": "Abuja",
    "Gombe": "Gombe",
    "Imo": "Owerri",
    "Jigawa": "Dutse",
    "Kaduna": "Kaduna",
    "Kano": "Kano",
    "Katsina": "Katsina",
    "Kebbi": "Birnin Kebbi",
    "Kogi": "Lokoja",
    "Kwara": "Ilorin",
    "Lagos": "Ikeja",
    "Nasarawa": "Lafia",
    "Niger": "Minna",
    "Ogun": "Abeokuta",
    "Ondo": "Akure",
    "Osun": "Osogbo",
    "Oyo": "Ibadan",
    "Plateau": "Jos",
    "Rivers": "Port Harcourt",
    "Sokoto": "Sokoto",
    "Taraba": "Jalingo",
    "Yobe": "Damaturu",
    "Zamfara": "Gusau",
}

STATE_NAMES: List[str] = sorted(STATES.keys())

# =============================================================================
# Local Government Areas (774 LGAs)
# Subset of major LGAs per state - full list would be 774 entries
# =============================================================================

LGAS: Dict[str, List[str]] = {
    "Abia": [
        "Aba North", "Aba South", "Arochukwu", "Bende", "Ikwuano",
        "Isiala Ngwa North", "Isiala Ngwa South", "Isuikwuato", "Obi Ngwa",
        "Ohafia", "Osisioma Ngwa", "Ugwunagbo", "Ukwa East", "Ukwa West",
        "Umuahia North", "Umuahia South", "Umu Nneochi",
    ],  # 17
    "Adamawa": [
        "Demsa", "Fufore", "Ganye", "Girei", "Gombi", "Guyuk", "Hong",
        "Jada", "Lamurde", "Madagali", "Maiha", "Mayo-Belwa", "Michika",
        "Mubi North", "Mubi South", "Numan", "Shelleng", "Song", "Toungo",
        "Yola North", "Yola South",
    ],  # 21
    "Akwa Ibom": [
        "Abak", "Eastern Obolo", "Eket", "Esit Eket", "Essien Udim",
        "Etim Ekpo", "Etinan", "Ibeno", "Ibesikpo Asutan", "Ibiono Ibom",
        "Ika", "Ikono", "Ikot Abasi", "Ikot Ekpene", "Ini", "Itu", "Mbo",
        "Mkpat Enin", "Nsit Atai", "Nsit Ibom", "Nsit Ubium", "Obot Akara",
        "Okobo", "Onna", "Oron", "Oruk Anam", "Udung Uko", "Ukanafun",
        "Uruan", "Urue-Offong/Oruko", "Uyo",
    ],  # 31
    "Anambra": [
        "Aguata", "Anambra East", "Anambra West", "Anaocha", "Awka North",
        "Awka South", "Ayamelum", "Dunukofia", "Ekwusigo", "Idemili North",
        "Idemili South", "Ihiala", "Njikoka", "Nnewi North", "Nnewi South",
        "Ogbaru", "Onitsha North", "Onitsha South", "Orumba North",
        "Orumba South", "Oyi",
    ],  # 21
    "Bauchi": [
        "Alkaleri", "Bauchi", "Bogoro", "Damban", "Darazo", "Dass",
        "Gamawa", "Ganjuwa", "Giade", "Itas/Gadau", "Jama'are", "Katagum",
        "Kirfi", "Misau", "Ningi", "Shira", "Tafawa Balewa", "Toro",
        "Warji", "Zaki",
    ],  # 20
    "Bayelsa": [
        "Brass", "Ekeremor", "Kolokuma/Opokuma", "Nembe", "Ogbia",
        "Sagbama", "Southern Ijaw", "Yenagoa",
    ],  # 8
    "Benue": [
        "Ado", "Agatu", "Apa", "Buruku", "Gboko", "Guma", "Gwer East",
        "Gwer West", "Katsina-Ala", "Konshisha", "Kwande", "Logo",
        "Makurdi", "Obi", "Ogbadibo", "Ohimini", "Oju", "Okpokwu",
        "Oturkpo", "Tarka", "Ukum", "Ushongo", "Vandeikya",
    ],  # 23
    "Borno": [
        "Abadam", "Askira/Uba", "Bama", "Bayo", "Biu", "Chibok",
        "Damboa", "Dikwa", "Gubio", "Guzamala", "Gwoza", "Hawul", "Jere",
        "Kaga", "Kala/Balge", "Konduga", "Kukawa", "Kwaya Kusar", "Mafa",
        "Magumeri", "Maiduguri", "Marte", "Mobbar", "Monguno", "Ngala",
        "Nganzai", "Shani",
    ],  # 27
    "Cross River": [
        "Abi", "Akamkpa", "Akpabuyo", "Bakassi", "Bekwarra", "Biase",
        "Boki", "Calabar Municipal", "Calabar South", "Etung", "Ikom",
        "Obanliku", "Obubra", "Obudu", "Odukpani", "Ogoja", "Yakurr",
        "Yala",
    ],  # 18
    "Delta": [
        "Aniocha North", "Aniocha South", "Bomadi", "Burutu", "Ethiope East",
        "Ethiope West", "Ika North East", "Ika South", "Isoko North",
        "Isoko South", "Ndokwa East", "Ndokwa West", "Okpe", "Oshimili North",
        "Oshimili South", "Patani", "Sapele", "Udu", "Ughelli North",
        "Ughelli South", "Ukwuani", "Uvwie", "Warri North", "Warri South",
        "Warri South West",
    ],  # 25
    "Ebonyi": [
        "Abakaliki", "Afikpo North", "Afikpo South", "Ebonyi", "Ezza North",
        "Ezza South", "Ikwo", "Ishielu", "Ivo", "Izzi", "Ohaozara",
        "Ohaukwu", "Onicha",
    ],  # 13
    "Edo": [
        "Akoko-Edo", "Egor", "Esan Central", "Esan North-East",
        "Esan South-East", "Esan West", "Etsako Central", "Etsako East",
        "Etsako West", "Igueben", "Ikpoba-Okha", "Oredo", "Orhionmwon",
        "Ovia North-East", "Ovia South-West", "Owan East", "Owan West",
        "Uhunmwonde",
    ],  # 18
    "Ekiti": [
        "Ado-Ekiti", "Efon", "Ekiti East", "Ekiti South-West", "Ekiti West",
        "Emure", "Gbonyin", "Ido-Osi", "Ijero", "Ikere", "Ikole",
        "Ilejemeje", "Irepodun/Ifelodun", "Ise/Orun", "Moba", "Oye",
    ],  # 16
    "Enugu": [
        "Aninri", "Awgu", "Enugu East", "Enugu North", "Enugu South",
        "Ezeagu", "Igbo-Etiti", "Igbo-Eze North", "Igbo-Eze South",
        "Isi-Uzo", "Nkanu East", "Nkanu West", "Nsukka", "Oji River",
        "Udenu", "Udi", "Uzo-Uwani",
    ],  # 17
    "FCT": [
        "Abaji", "Abuja Municipal", "Bwari", "Gwagwalada", "Kuje", "Kwali",
    ],  # 6
    "Gombe": [
        "Akko", "Balanga", "Billiri", "Dukku", "Funakaye", "Gombe",
        "Kaltungo", "Kwami", "Nafada", "Shongom", "Yamaltu/Deba",
    ],  # 11
    "Imo": [
        "Aboh Mbaise", "Ahiazu Mbaise", "Ehime Mbano", "Ezinihitte",
        "Ideato North", "Ideato South", "Ihitte/Uboma", "Ikeduru",
        "Isiala Mbano", "Isu", "Mbaitoli", "Ngor-Okpala", "Njaba",
        "Nkwerre", "Nwangele", "Obowo", "Oguta", "Ohaji/Egbema",
        "Okigwe", "Onuimo", "Orlu", "Orsu", "Oru East", "Oru West",
        "Owerri Municipal", "Owerri North", "Owerri West",
    ],  # 27
    "Jigawa": [
        "Auyo", "Babura", "Biriniwa", "Birnin Kudu", "Buji", "Dutse",
        "Gagarawa", "Garki", "Gumel", "Guri", "Gwaram", "Gwiwa",
        "Hadejia", "Jahun", "Kafin Hausa", "Kaugama", "Kazaure",
        "Kiri Kasama", "Kiyawa", "Maigatari", "Malam Madori", "Miga",
        "Ringim", "Roni", "Sule Tankarkar", "Taura", "Yankwashi",
    ],  # 27
    "Kaduna": [
        "Birnin Gwari", "Chikun", "Giwa", "Igabi", "Ikara", "Jaba",
        "Jema'a", "Kachia", "Kaduna North", "Kaduna South", "Kagarko",
        "Kajuru", "Kaura", "Kauru", "Kubau", "Kudan", "Lere", "Makarfi",
        "Sabon Gari", "Sanga", "Soba", "Zangon Kataf", "Zaria",
    ],  # 23
    "Kano": [
        "Ajingi", "Albasu", "Bagwai", "Bebeji", "Bichi", "Bunkure",
        "Dala", "Dambatta", "Dawakin Kudu", "Dawakin Tofa", "Doguwa",
        "Fagge", "Gabasawa", "Garko", "Garun Mallam", "Gaya", "Gezawa",
        "Gwale", "Gwarzo", "Kabo", "Kano Municipal", "Karaye", "Kibiya",
        "Kiru", "Kumbotso", "Kunchi", "Kura", "Madobi", "Makoda",
        "Minjibir", "Nasarawa", "Rano", "Rimin Gado", "Rogo", "Shanono",
        "Sumaila", "Takai", "Tarauni", "Tofa", "Tsanyawa", "Tudun Wada",
        "Ungogo", "Warawa", "Wudil",
    ],  # 44
    "Katsina": [
        "Bakori", "Batagarawa", "Batsari", "Baure", "Bindawa", "Charanchi",
        "Dan Musa", "Dandume", "Danja", "Daura", "Dutsi", "Dutsin-Ma",
        "Faskari", "Funtua", "Ingawa", "Jibia", "Kafur", "Kaita",
        "Kankara", "Kankia", "Katsina", "Kurfi", "Kusada", "Mai'Adua",
        "Malumfashi", "Mani", "Mashi", "Matazu", "Musawa", "Rimi",
        "Sabuwa", "Safana", "Sandamu", "Zango",
    ],  # 34
    "Kebbi": [
        "Aleiro", "Arewa Dandi", "Argungu", "Augie", "Bagudo",
        "Birnin Kebbi", "Bunza", "Dandi", "Fakai", "Gwandu", "Jega",
        "Kalgo", "Koko/Besse", "Maiyama", "Ngaski", "Sakaba", "Shanga",
        "Suru", "Wasagu/Danko", "Yauri", "Zuru",
    ],  # 21
    "Kogi": [
        "Adavi", "Ajaokuta", "Ankpa", "Bassa", "Dekina", "Ibaji", "Idah",
        "Igalamela-Odolu", "Ijumu", "Kabba/Bunu", "Kogi", "Lokoja",
        "Mopa-Muro", "Ofu", "Ogori/Magongo", "Okehi", "Okene",
        "Olamaboro", "Omala", "Yagba East", "Yagba West",
    ],  # 21
    "Kwara": [
        "Asa", "Baruten", "Edu", "Ekiti", "Ifelodun", "Ilorin East",
        "Ilorin South", "Ilorin West", "Irepodun", "Isin", "Kaiama",
        "Moro", "Offa", "Oke Ero", "Oyun", "Pategi",
    ],  # 16
    "Lagos": [
        "Agege", "Ajeromi-Ifelodun", "Alimosho", "Amuwo-Odofin", "Apapa",
        "Badagry", "Epe", "Eti-Osa", "Ibeju-Lekki", "Ifako-Ijaiye",
        "Ikeja", "Ikorodu", "Kosofe", "Lagos Island", "Lagos Mainland",
        "Mushin", "Ojo", "Oshodi-Isolo", "Shomolu", "Surulere",
    ],  # 20
    "Nasarawa": [
        "Akwanga", "Awe", "Doma", "Karu", "Keana", "Keffi", "Kokona",
        "Lafia", "Nasarawa", "Nasarawa Eggon", "Obi", "Toto", "Wamba",
    ],  # 13
    "Niger": [
        "Agaie", "Agwara", "Bida", "Borgu", "Bosso", "Chanchaga",
        "Edati", "Gbako", "Gurara", "Katcha", "Kontagora", "Lapai",
        "Lavun", "Magama", "Mariga", "Mashegu", "Mokwa", "Muya",
        "Paikoro", "Rafi", "Rijau", "Shiroro", "Suleja", "Tafa",
        "Wushishi",
    ],  # 25
    "Ogun": [
        "Abeokuta North", "Abeokuta South", "Ado-Odo/Ota", "Ewekoro",
        "Ifo", "Ijebu East", "Ijebu North", "Ijebu North East", "Ijebu Ode",
        "Ikenne", "Imeko Afon", "Ipokia", "Obafemi Owode", "Odeda",
        "Odogbolu", "Ogun Waterside", "Remo North", "Sagamu",
        "Yewa North", "Yewa South",
    ],  # 20
    "Ondo": [
        "Akoko North-East", "Akoko North-West", "Akoko South-East",
        "Akoko South-West", "Akure North", "Akure South", "Ese Odo",
        "Idanre", "Ifedore", "Ilaje", "Ile Oluji/Okeigbo", "Irele",
        "Odigbo", "Okitipupa", "Ondo East", "Ondo West", "Ose", "Owo",
    ],  # 18
    "Osun": [
        "Aiyedade", "Aiyedire", "Atakumosa East", "Atakumosa West",
        "Boluwaduro", "Boripe", "Ede North", "Ede South", "Egbedore",
        "Ejigbo", "Ife Central", "Ife East", "Ife North", "Ife South",
        "Ifedayo", "Ifelodun", "Ila", "Ilesa East", "Ilesa West",
        "Irepodun", "Irewole", "Isokan", "Iwo", "Obokun", "Odo-Otin",
        "Ola Oluwa", "Olorunda", "Oriade", "Orolu", "Osogbo",
    ],  # 30
    "Oyo": [
        "Afijio", "Akinyele", "Atiba", "Atisbo", "Egbeda",
        "Ibadan North", "Ibadan North-East", "Ibadan North-West",
        "Ibadan South-East", "Ibadan South-West", "Ibarapa Central",
        "Ibarapa East", "Ibarapa North", "Ido", "Irepo", "Iseyin",
        "Itesiwaju", "Iwajowa", "Kajola", "Lagelu", "Ogbomosho North",
        "Ogbomosho South", "Ogo Oluwa", "Olorunsogo", "Oluyole",
        "Ona Ara", "Orelope", "Ori Ire", "Oyo East", "Oyo West",
        "Saki East", "Saki West", "Surulere",
    ],  # 33
    "Plateau": [
        "Barkin Ladi", "Bassa", "Bokkos", "Jos East", "Jos North",
        "Jos South", "Kanam", "Kanke", "Langtang North", "Langtang South",
        "Mangu", "Mikang", "Pankshin", "Qua'an Pan", "Riyom", "Shendam",
        "Wase",
    ],  # 17
    "Rivers": [
        "Abua/Odual", "Ahoada East", "Ahoada West", "Akuku-Toru", "Andoni",
        "Asari-Toru", "Bonny", "Degema", "Eleme", "Emohua", "Etche",
        "Gokana", "Ikwerre", "Khana", "Obio/Akpor", "Ogba/Egbema/Ndoni",
        "Ogu/Bolo", "Okrika", "Omuma", "Opobo/Nkoro", "Oyigbo",
        "Port Harcourt", "Tai",
    ],  # 23
    "Sokoto": [
        "Binji", "Bodinga", "Dange Shuni", "Gada", "Goronyo", "Gudu",
        "Gwadabawa", "Illela", "Isa", "Kebbe", "Kware", "Rabah",
        "Sabon Birni", "Shagari", "Silame", "Sokoto North", "Sokoto South",
        "Tambuwal", "Tangaza", "Tureta", "Wamako", "Wurno", "Yabo",
    ],  # 23
    "Taraba": [
        "Ardo Kola", "Bali", "Donga", "Gashaka", "Gassol", "Ibi",
        "Jalingo", "Karim Lamido", "Kurmi", "Lau", "Sardauna", "Takum",
        "Ussa", "Wukari", "Yorro", "Zing",
    ],  # 16
    "Yobe": [
        "Bade", "Bursari", "Damaturu", "Fika", "Fune", "Geidam", "Gujba",
        "Gulani", "Jakusko", "Karasuwa", "Machina", "Nangere", "Nguru",
        "Potiskum", "Tarmuwa", "Yunusari", "Yusufari",
    ],  # 17
    "Zamfara": [
        "Anka", "Bakura", "Birnin Magaji/Kiyaw", "Bukkuyum", "Bungudu",
        "Gummi", "Gusau", "Kaura Namoda", "Maradun", "Maru", "Shinkafi",
        "Talata Mafara", "Tsafe", "Zurmi",
    ],  # 14
}

# =============================================================================
# Nigerian Banks
# =============================================================================

BANKS: Dict[str, str] = {
    # Traditional banks
    "Access Bank": "044",
    "Citibank": "023",
    "Ecobank": "050",
    "Fidelity Bank": "070",
    "First Bank": "011",
    "First City Monument Bank": "214",
    "Globus Bank": "103",
    "Guaranty Trust Bank": "058",
    "Heritage Bank": "030",
    "Keystone Bank": "082",
    "Polaris Bank": "076",
    "Providus Bank": "101",
    "Stanbic IBTC": "221",
    "Standard Chartered": "068",
    "Sterling Bank": "232",
    "SunTrust Bank": "100",
    "Titan Trust Bank": "102",
    "Union Bank": "032",
    "United Bank for Africa": "033",
    "Unity Bank": "215",
    "Wema Bank": "035",
    "Zenith Bank": "057",
    # Digital/Fintech banks
    "Kuda Bank": "090267",
    "OPay": "100004",
    "PalmPay": "100033",
    "Moniepoint": "100022",
    "Carbon": "100026",
    "Sparkle": "100269",
    "VFD Microfinance Bank": "090110",
}

BANK_NAMES: List[str] = sorted(BANKS.keys())

# =============================================================================
# Telecom Operators
# =============================================================================

TELCOS: Dict[str, Dict[str, any]] = {
    "MTN": {
        "prefixes": [
            "0703", "0706", "0803", "0806", "0810", "0813", "0814", "0816",
            "0903", "0906", "0913", "0916",
        ],
    },
    "Airtel": {
        "prefixes": [
            "0701", "0702", "0708", "0802", "0808", "0812",
            "0901", "0902", "0904", "0907", "0912",
        ],
    },
    "Glo": {
        "prefixes": [
            "0705", "0805", "0807", "0811", "0815", "0905", "0915",
        ],
    },
    "9mobile": {
        "prefixes": [
            "0809", "0817", "0818", "0908", "0909",
        ],
    },
}

TELCO_NAMES: List[str] = sorted(TELCOS.keys())

# =============================================================================
# Regex Patterns
# =============================================================================

# Nigerian phone number: 11 digits starting with 0, or with +234/234 prefix
PHONE_PATTERN = re.compile(
    r"^(?:0|\+?234)"  # Start with 0, 234, or +234
    r"[789][01]\d"     # Second digit 7/8/9, third digit 0/1, fourth any digit
    r"\d{7}$"          # Remaining 7 digits
)

# More permissive pattern for finding phones in text (allows spaces, dashes)
PHONE_PATTERN_LOOSE = re.compile(
    r"(?:0|\+?234)[\s\-]?"
    r"[789][01]\d[\s\-]?"
    r"\d{3}[\s\-]?"
    r"\d{4}"
)

# Bank Verification Number (BVN): 11 digits starting with 22
BVN_PATTERN = re.compile(r"^22\d{9}$")

# National Identification Number (NIN): 11 digits
NIN_PATTERN = re.compile(r"^\d{11}$")

# Nigerian bank account number: 10 digits (NUBAN)
NUBAN_PATTERN = re.compile(r"^\d{10}$")

# Naira amounts in text: ₦1,000 or N1000 or NGN 1,000
NAIRA_PATTERN = re.compile(
    r"(?:₦|NGN|N)\s?"
    r"[\d,]+(?:\.\d{2})?"
)

# =============================================================================
# Utility Functions
# =============================================================================

def format_naira(amount: float, include_kobo: bool = True) -> str:
    """Format a number as Nigerian Naira currency.

    Args:
        amount: The amount to format.
        include_kobo: Whether to include decimal places (kobo).

    Returns:
        Formatted string like "₦1,500,000.00".

    Example:
        >>> format_naira(1500000)
        '₦1,500,000.00'
        >>> format_naira(1500000, include_kobo=False)
        '₦1,500,000'
    """
    if include_kobo:
        return "₦{:,.2f}".format(amount)
    return "₦{:,}".format(int(amount))


def parse_naira(text: str) -> Optional[float]:
    """Parse a Naira amount from text.

    Args:
        text: String containing a Naira amount (e.g., "₦1,500,000.00").

    Returns:
        Float value, or None if parsing fails.

    Example:
        >>> parse_naira("₦1,500,000.00")
        1500000.0
        >>> parse_naira("NGN 50,000")
        50000.0
    """
    cleaned = re.sub(r"[₦NGN\s,]", "", text)
    try:
        return float(cleaned)
    except ValueError:
        return None


def is_valid_phone(phone: str) -> bool:
    """Check if a string is a valid Nigerian phone number.

    Args:
        phone: Phone number string to validate.

    Returns:
        True if valid Nigerian phone number format.

    Example:
        >>> is_valid_phone("08012345678")
        True
        >>> is_valid_phone("+2348012345678")
        True
        >>> is_valid_phone("12345")
        False
    """
    cleaned = re.sub(r"[\s\-]", "", phone)
    return bool(PHONE_PATTERN.match(cleaned))


def normalize_phone(phone: str) -> Optional[str]:
    """Normalize a Nigerian phone number to international format.

    Args:
        phone: Phone number in any common Nigerian format.

    Returns:
        Phone number in +234XXXXXXXXXX format, or None if invalid.

    Example:
        >>> normalize_phone("08012345678")
        '+2348012345678'
        >>> normalize_phone("234-801-234-5678")
        '+2348012345678'
    """
    cleaned = re.sub(r"[\s\-]", "", phone)
    if not PHONE_PATTERN.match(cleaned):
        return None
    if cleaned.startswith("+234"):
        return cleaned
    if cleaned.startswith("234"):
        return "+" + cleaned
    if cleaned.startswith("0"):
        return "+234" + cleaned[1:]
    return None


def get_telco(phone: str) -> Optional[str]:
    """Identify the telecom operator from a Nigerian phone number.

    Args:
        phone: Nigerian phone number.

    Returns:
        Telco name ('MTN', 'Airtel', 'Glo', '9mobile') or None.

    Example:
        >>> get_telco("08031234567")
        'MTN'
        >>> get_telco("08021234567")
        'Airtel'
    """
    cleaned = re.sub(r"[\s\-]", "", phone)
    # Normalize to 0xxx format for prefix matching
    if cleaned.startswith("+234"):
        cleaned = "0" + cleaned[4:]
    elif cleaned.startswith("234"):
        cleaned = "0" + cleaned[3:]

    prefix = cleaned[:4]
    for telco, info in TELCOS.items():
        if prefix in info["prefixes"]:
            return telco
    return None


def is_valid_bvn(bvn: str) -> bool:
    """Check if a string is a valid BVN format.

    Args:
        bvn: Bank Verification Number to validate.

    Returns:
        True if valid BVN format (11 digits starting with 22).
    """
    return bool(BVN_PATTERN.match(bvn))


def is_valid_nin(nin: str) -> bool:
    """Check if a string is a valid NIN format.

    Args:
        nin: National Identification Number to validate.

    Returns:
        True if valid NIN format (11 digits).
    """
    return bool(NIN_PATTERN.match(nin))


# =============================================================================
# Nigerian Pidgin (Naija) Particles and Common Words
# =============================================================================

# These are discourse particles, intensifiers, and function words that are
# unique to Nigerian Pidgin and should be preserved during text cleaning.
# Standard NLP tools often strip these as "noise" or "errors".

PIDGIN_PARTICLES: set = {
    # Discourse particles (sentence modifiers)
    "sha",       # anyway, though (softener)
    "sef",       # even, self (emphasis)
    "abeg",      # please (politeness marker)
    "abi",       # or, right? (question tag)
    "shey",      # isn't it? (question tag)
    "shebi",     # right? (confirmation)
    "na",        # copula/focus marker
    "dey",       # progressive marker / to be
    "no",        # negation
    "wey",       # relative pronoun (that/which)
    "oya",       # let's go, come on
    "jare",      # please (Yoruba origin)
    "jor",       # please (variant)
    "biko",      # please (Igbo origin)
    "walahi",    # I swear (Hausa origin)
    "wallahi",   # I swear (variant)

    # Intensifiers and modifiers
    "well",      # very, really
    "die",       # extremely (to die for)
    "gidigba",   # seriously, plenty
    "scatter",   # extremely
    "proper",    # properly, really
    "sharp",     # quickly
    "quick",     # quickly

    # Common Pidgin verbs/words often stripped
    "dey",       # be/is (progressive)
    "chop",      # eat
    "gist",      # chat/gossip
    "vex",       # angry
    "japa",      # run away/emigrate
    "sabi",      # know
    "wan",       # want
    "go",        # future marker
    "don",       # perfect marker
    "fit",       # can/able

    # Pronouns and determiners
    "wetin",     # what
    "weda",      # whether
    "una",       # you (plural)
    "dem",       # them/they
    "am",        # him/her/it
    "im",        # his/her

    # Greetings and expressions
    "howfar",    # hello, how are you
    "ehen",      # I see, okay
    "ehn",       # really? (question)
    "kai",       # exclamation
    "chei",      # exclamation
    "wahala",    # trouble
    "palava",    # problem
    "yawa",      # trouble
    "kolo",      # crazy
    "mumu",      # fool
    "oga",       # boss
    "madam",     # ma'am
    "bros",      # brother
    "sista",     # sister
    "pikin",     # child
    "baba",      # father/old man
    "mama",      # mother
}

# Common Pidgin multi-word expressions
PIDGIN_EXPRESSIONS: set = {
    "how far",
    "no wahala",
    "no vex",
    "e go be",
    "na so",
    "na wa",
    "no be",
    "wey dey",
    "for where",
    "make we",
    "i no",
    "you no",
    "e no",
    "dem no",
    "no dey",
}



#!/usr/bin/env python3
"""Tokenizer benchmark: measures how major AI tokenizers handle Nigerian languages vs English.

Compares GPT-4, GPT-4o, Llama 3, Gemma 2, Mistral, BERT multilingual, and XLM-RoBERTa
tokenizers across Yoruba, Hausa, Igbo, Nigerian Pidgin, and English.

Usage:
    python benchmarks/tokenizer_benchmark.py           # full benchmark (100 entries)
    python benchmarks/tokenizer_benchmark.py --quick    # quick mode (10 entries)

Outputs:
    benchmarks/results/benchmark_results.json    - raw data
    benchmarks/results/benchmark_summary.csv     - aggregate stats
"""

from __future__ import annotations

import argparse
import csv
import json
import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Tuple

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)

# --- Constants ---

LANGUAGES = ["eng", "yor", "hau", "ibo", "pcm"]
LANG_NAMES = {
    "eng": "English",
    "yor": "Yoruba",
    "hau": "Hausa",
    "ibo": "Igbo",
    "pcm": "Pidgin",
}

# GPT-4o pricing: $2.50 per 1M input tokens
GPT4O_USD_PER_1M_TOKENS = 2.50
USD_TO_NGN = 1350.0  # CBN official rate, Feb 2026

TOKENIZER_IDS = [
    "gpt4_cl100k",
    "gpt4o_o200k",
    "llama3",
    "gemma",
    "mistral",
    "bert_multilingual",
    "xlm_roberta",
]

TOKENIZER_NAMES = {
    "gpt4_cl100k": "GPT-4 (cl100k_base)",
    "gpt4o_o200k": "GPT-4o (o200k_base)",
    "llama3": "Llama 3",
    "gemma": "Gemma 2",
    "mistral": "Mistral v0.3",
    "bert_multilingual": "BERT Multilingual",
    "xlm_roberta": "XLM-RoBERTa",
}


# --- Tokenizer Loading ---


def load_tiktoken(encoding_name: str):
    """Load a tiktoken encoding."""
    import tiktoken

    return tiktoken.get_encoding(encoding_name)


def load_hf_tokenizer(model_name: str):
    """Load a HuggingFace tokenizer."""
    from transformers import AutoTokenizer

    return AutoTokenizer.from_pretrained(model_name)


LLAMA3_CANDIDATES = [
    "unsloth/llama-3-8b",
    "NousResearch/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B",
]


def _load_llama3_tokenizer():
    """Try multiple Llama 3 tokenizer sources (gated model has mirrors)."""
    from transformers import AutoTokenizer

    for model_id in LLAMA3_CANDIDATES:
        try:
            return AutoTokenizer.from_pretrained(model_id)
        except (OSError, Exception) as e:
            logger.info(f"  Could not load from {model_id}: {type(e).__name__}")
            continue
    raise RuntimeError(
        "Could not load Llama 3 tokenizer. Try: huggingface-cli login "
        "(requires access to meta-llama/Meta-Llama-3-8B)"
    )


def get_tokenizers() -> Dict[str, Any]:
    """Load all tokenizers. Returns dict mapping tokenizer_id -> tokenizer object.

    Skips tokenizers that fail to load (with a warning) rather than crashing.
    """
    tokenizers = {}

    specs = [
        ("gpt4_cl100k", "tiktoken", "cl100k_base"),
        ("gpt4o_o200k", "tiktoken", "o200k_base"),
        ("llama3", "llama3", None),
        ("gemma", "hf", "google/gemma-2b"),
        ("mistral", "hf", "mistralai/Mistral-7B-v0.3"),
        ("bert_multilingual", "hf", "bert-base-multilingual-cased"),
        ("xlm_roberta", "hf", "xlm-roberta-base"),
    ]

    for tok_id, tok_type, model_id in specs:
        name = TOKENIZER_NAMES[tok_id]
        logger.info(f"Loading {name}...")
        try:
            if tok_type == "tiktoken":
                tokenizers[tok_id] = ("tiktoken", load_tiktoken(model_id))
            elif tok_type == "llama3":
                tokenizers[tok_id] = ("hf", _load_llama3_tokenizer())
            else:
                tokenizers[tok_id] = ("hf", load_hf_tokenizer(model_id))
        except Exception as e:
            logger.warning(f"  SKIPPED {name}: {e}\n")
            continue

    loaded = [TOKENIZER_NAMES[t] for t in tokenizers]
    logger.info(f"Loaded {len(tokenizers)} tokenizers: {', '.join(loaded)}\n")
    return tokenizers


def tokenize(text: str, tokenizer_type: str, tokenizer: Any) -> Tuple[List[int], List[str]]:
    """Tokenize text and return (token_ids, token_strings)."""
    if tokenizer_type == "tiktoken":
        token_ids = tokenizer.encode(text)
        token_strings = [tokenizer.decode([tid]) for tid in token_ids]
    else:
        encoded = tokenizer(text, add_special_tokens=False)
        token_ids = encoded["input_ids"]
        token_strings = tokenizer.convert_ids_to_tokens(token_ids)
    return token_ids, token_strings


# --- Data Loading ---


def load_benchmark_words(data_path: Path, quick: bool = False) -> List[Dict[str, Any]]:
    """Load benchmark words from JSON file."""
    with open(data_path, "r", encoding="utf-8") as f:
        data = json.load(f)

    entries = data["entries"]

    if quick:
        # Take first 10 entries for quick mode
        entries = entries[:10]
        logger.info(f"Quick mode: using {len(entries)} entries per language\n")
    else:
        logger.info(f"Full mode: using {len(entries)} entries per language\n")

    return entries


# --- Benchmarking ---


def run_benchmark(
    entries: List[Dict[str, Any]], tokenizers: Dict[str, Any]
) -> List[Dict[str, Any]]:
    """Run the benchmark across all entries and tokenizers.

    Returns a list of result dicts, one per (entry, language, tokenizer) combo.
    """
    results = []

    for entry in entries:
        entry_id = entry["id"]
        category = entry["category"]

        # Get English token counts first (for ratio calculation)
        eng_text = entry["eng"]
        eng_token_counts = {}
        for tok_id, (tok_type, tok_obj) in tokenizers.items():
            ids, _ = tokenize(eng_text, tok_type, tok_obj)
            eng_token_counts[tok_id] = len(ids)

        for lang in LANGUAGES:
            text = entry[lang]

            for tok_id, (tok_type, tok_obj) in tokenizers.items():
                token_ids, token_strings = tokenize(text, tok_type, tok_obj)
                n_tokens = len(token_ids)
                eng_count = eng_token_counts[tok_id]
                ratio = n_tokens / eng_count if eng_count > 0 else float("inf")

                results.append(
                    {
                        "entry_id": entry_id,
                        "category": category,
                        "language": lang,
                        "language_name": LANG_NAMES[lang],
                        "text": text,
                        "tokenizer": tok_id,
                        "tokenizer_name": TOKENIZER_NAMES[tok_id],
                        "n_tokens": n_tokens,
                        "token_splits": token_strings,
                        "english_equivalent": eng_text,
                        "english_tokens": eng_count,
                        "token_ratio_vs_english": round(ratio, 3),
                    }
                )

    return results


def compute_aggregate_stats(results: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Compute aggregate statistics per language per tokenizer."""
    import numpy as np

    # Group results by (language, tokenizer)
    groups: Dict[Tuple[str, str], List[Dict[str, Any]]] = {}
    for r in results:
        key = (r["language"], r["tokenizer"])
        groups.setdefault(key, []).append(r)

    stats = []
    for (lang, tok_id), group in sorted(groups.items()):
        token_counts = [r["n_tokens"] for r in group]
        ratios = [r["token_ratio_vs_english"] for r in group]

        mean_tokens = float(np.mean(token_counts))
        mean_ratio = float(np.mean(ratios))

        # Cost multiplier: ratio of mean tokens to English mean tokens
        eng_key = ("eng", tok_id)
        if eng_key in groups:
            eng_mean = float(np.mean([r["n_tokens"] for r in groups[eng_key]]))
            cost_multiplier = mean_tokens / eng_mean if eng_mean > 0 else 1.0
        else:
            cost_multiplier = mean_ratio

        # Worst offender: highest token ratio in this group
        worst = max(group, key=lambda r: r["token_ratio_vs_english"])

        # Naira cost per 1M tokens equivalent
        naira_per_1m = GPT4O_USD_PER_1M_TOKENS * USD_TO_NGN  # base cost in NGN
        estimated_naira_cost = naira_per_1m * cost_multiplier

        stats.append(
            {
                "language": lang,
                "language_name": LANG_NAMES[lang],
                "tokenizer": tok_id,
                "tokenizer_name": TOKENIZER_NAMES[tok_id],
                "mean_tokens_per_entry": round(mean_tokens, 2),
                "mean_token_ratio_vs_english": round(mean_ratio, 3),
                "cost_multiplier_vs_english": round(cost_multiplier, 3),
                "estimated_naira_per_1m_equiv": round(estimated_naira_cost, 0),
                "worst_word": worst["text"],
                "worst_word_tokens": worst["n_tokens"],
                "worst_word_ratio": worst["token_ratio_vs_english"],
            }
        )

    return stats


def find_top_worst_tokenized(results: List[Dict[str, Any]], top_n: int = 10) -> List[Dict[str, Any]]:
    """Find the top N worst tokenized words across all languages (excluding English)."""
    non_eng = [r for r in results if r["language"] != "eng"]
    sorted_results = sorted(non_eng, key=lambda r: r["token_ratio_vs_english"], reverse=True)

    seen = set()
    top = []
    for r in sorted_results:
        # Deduplicate by (text, tokenizer)
        key = (r["text"], r["tokenizer"])
        if key in seen:
            continue
        seen.add(key)
        top.append(
            {
                "text": r["text"],
                "language": r["language_name"],
                "tokenizer": r["tokenizer_name"],
                "n_tokens": r["n_tokens"],
                "english_equivalent": r["english_equivalent"],
                "english_tokens": r["english_tokens"],
                "ratio": r["token_ratio_vs_english"],
                "token_splits": r["token_splits"],
            }
        )
        if len(top) >= top_n:
            break

    return top


# --- Output ---


def save_results(
    results: List[Dict[str, Any]],
    stats: List[Dict[str, Any]],
    top_worst: List[Dict[str, Any]],
    output_dir: Path,
) -> None:
    """Save all results to JSON and CSV files."""
    output_dir.mkdir(parents=True, exist_ok=True)

    # Raw results JSON
    results_path = output_dir / "benchmark_results.json"
    with open(results_path, "w", encoding="utf-8") as f:
        json.dump(
            {
                "metadata": {
                    "tokenizers": TOKENIZER_NAMES,
                    "languages": LANG_NAMES,
                    "gpt4o_pricing_usd_per_1m": GPT4O_USD_PER_1M_TOKENS,
                    "usd_to_ngn_rate": USD_TO_NGN,
                },
                "results": results,
                "aggregate_stats": stats,
                "top_worst_tokenized": top_worst,
            },
            f,
            ensure_ascii=False,
            indent=2,
        )
    logger.info(f"Saved raw results to {results_path}")

    # Summary CSV
    csv_path = output_dir / "benchmark_summary.csv"
    fieldnames = [
        "language",
        "language_name",
        "tokenizer",
        "tokenizer_name",
        "mean_tokens_per_entry",
        "mean_token_ratio_vs_english",
        "cost_multiplier_vs_english",
        "estimated_naira_per_1m_equiv",
        "worst_word",
        "worst_word_tokens",
        "worst_word_ratio",
    ]
    with open(csv_path, "w", encoding="utf-8", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(stats)
    logger.info(f"Saved summary CSV to {csv_path}")


def print_summary_table(stats: List[Dict[str, Any]], top_worst: List[Dict[str, Any]]) -> None:
    """Print a clean summary table to stdout."""
    print("\n" + "=" * 90)
    print("  TOKENIZER BENCHMARK: Nigerian Languages vs English")
    print("=" * 90)

    # Group by tokenizer (use TOKENIZER_IDS order for those present)
    present_tok_ids = [t for t in TOKENIZER_IDS if any(s["tokenizer"] == t for s in stats)]
    for tok_id in present_tok_ids:
        tok_name = TOKENIZER_NAMES[tok_id]
        tok_stats = [s for s in stats if s["tokenizer"] == tok_id]
        if not tok_stats:
            continue

        print(f"\n{'─' * 90}")
        print(f"  {tok_name}")
        print(f"{'─' * 90}")
        print(f"  {'Language':<12} {'Mean Tokens':>12} {'Ratio vs Eng':>14} {'Cost Mult':>11} {'Est. ₦/1M':>12}")
        print(f"  {'─' * 10:<12} {'─' * 10:>12} {'─' * 12:>14} {'─' * 9:>11} {'─' * 10:>12}")

        for s in tok_stats:
            lang_name = s["language_name"]
            mean_tok = s["mean_tokens_per_entry"]
            ratio = s["mean_token_ratio_vs_english"]
            cost_mult = s["cost_multiplier_vs_english"]
            naira = s["estimated_naira_per_1m_equiv"]
            print(
                f"  {lang_name:<12} {mean_tok:>12.1f} {ratio:>14.2f}x {cost_mult:>10.2f}x ₦{naira:>10,.0f}"
            )

    # Top worst
    print(f"\n{'=' * 90}")
    print("  TOP 10 WORST TOKENIZED (highest token explosion vs English)")
    print(f"{'=' * 90}")
    print(f"  {'#':<3} {'Text':<30} {'Lang':<8} {'Tokenizer':<22} {'Tokens':>7} {'Ratio':>7}")
    print(f"  {'─' * 2:<3} {'─' * 28:<30} {'─' * 6:<8} {'─' * 20:<22} {'─' * 5:>7} {'─' * 5:>7}")

    for i, w in enumerate(top_worst, 1):
        text = w["text"][:28]
        print(
            f"  {i:<3} {text:<30} {w['language']:<8} {w['tokenizer']:<22} {w['n_tokens']:>7} {w['ratio']:>6.1f}x"
        )

    # Key takeaway
    print(f"\n{'=' * 90}")
    print("  KEY INSIGHT")
    print(f"{'=' * 90}")

    # Find language with highest average cost across tokenizers
    lang_avg_cost: Dict[str, List[float]] = {}
    for s in stats:
        if s["language"] != "eng":
            lang_avg_cost.setdefault(s["language_name"], []).append(
                s["cost_multiplier_vs_english"]
            )

    if lang_avg_cost:
        import numpy as np

        most_expensive = max(lang_avg_cost.items(), key=lambda x: np.mean(x[1]))
        avg_mult = np.mean(most_expensive[1])
        print(
            f"  {most_expensive[0]} is the most expensive language to process,"
        )
        print(
            f"  costing on average {avg_mult:.2f}x more tokens than English."
        )
        base_naira = GPT4O_USD_PER_1M_TOKENS * USD_TO_NGN
        print(
            f"  If English costs ₦{base_naira:,.0f} per 1M tokens (GPT-4o),"
        )
        print(
            f"  {most_expensive[0]} costs ~₦{base_naira * avg_mult:,.0f} for equivalent content."
        )

    print(f"\n{'=' * 90}\n")


# --- Main ---


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Benchmark AI tokenizers on Nigerian languages vs English"
    )
    parser.add_argument(
        "--quick",
        action="store_true",
        help="Run on 10 words per language for fast testing",
    )
    parser.add_argument(
        "--data",
        type=str,
        default=None,
        help="Path to benchmark_words.json (auto-detected if not set)",
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Output directory for results (default: results/)",
    )
    args = parser.parse_args()

    # Find paths
    script_dir = Path(__file__).resolve().parent

    if args.data:
        data_path = Path(args.data)
    else:
        data_path = script_dir / "benchmark_words.json"

    if args.output:
        output_dir = Path(args.output)
    else:
        output_dir = script_dir / "results"

    if not data_path.exists():
        logger.error(f"Benchmark data not found at {data_path}")
        sys.exit(1)

    # Load data
    entries = load_benchmark_words(data_path, quick=args.quick)

    # Load tokenizers
    tokenizers = get_tokenizers()

    # Run benchmark
    logger.info("Running benchmark...")
    results = run_benchmark(entries, tokenizers)
    logger.info(f"Generated {len(results)} result entries.\n")

    # Compute stats
    stats = compute_aggregate_stats(results)
    top_worst = find_top_worst_tokenized(results, top_n=10)

    # Save outputs
    save_results(results, stats, top_worst, output_dir)

    # Print summary
    print_summary_table(stats, top_worst)

    logger.info("Benchmark complete!")


if __name__ == "__main__":
    main()



"""Tests for sentiment analysis module."""
from __future__ import annotations

import pytest


class TestAnalyzeSentiment:
    """Test analyze_sentiment function."""

    def test_positive_english(self):
        from naijaml.nlp import analyze_sentiment
        result = analyze_sentiment("This is great, I love it!")
        assert result["label"] in {"positive", "negative", "neutral"}
        assert "confidence" in result
        assert "scores" in result
        assert 0.0 <= result["confidence"] <= 1.0

    def test_negative_pidgin(self):
        from naijaml.nlp import analyze_sentiment
        result = analyze_sentiment("I no like am at all, e bad")
        assert result["label"] in {"positive", "negative", "neutral"}

    def test_result_structure(self):
        from naijaml.nlp import analyze_sentiment
        result = analyze_sentiment("Test text")
        assert "label" in result
        assert "confidence" in result
        assert "scores" in result
        assert isinstance(result["scores"], dict)
        # Scores should sum to ~1.0
        total = sum(result["scores"].values())
        assert 0.99 <= total <= 1.01

    def test_empty_returns_neutral(self):
        from naijaml.nlp import analyze_sentiment
        result = analyze_sentiment("")
        assert result["label"] == "neutral"


class TestGetSentiment:
    """Test get_sentiment simplified API."""

    def test_returns_string(self):
        from naijaml.nlp import get_sentiment
        result = get_sentiment("This is good")
        assert isinstance(result, str)
        assert result in {"positive", "negative", "neutral"}


class TestGetSentimentWithConfidence:
    """Test get_sentiment_with_confidence."""

    def test_returns_tuple(self):
        from naijaml.nlp import get_sentiment_with_confidence
        label, conf = get_sentiment_with_confidence("Test text")
        assert isinstance(label, str)
        assert isinstance(conf, float)
        assert label in {"positive", "negative", "neutral"}
        assert 0.0 <= conf <= 1.0


class TestAnalyzeBatch:
    """Test batch sentiment analysis."""

    def test_batch_returns_list(self):
        from naijaml.nlp import analyze_sentiment_batch
        texts = ["good", "bad", "okay"]
        results = analyze_sentiment_batch(texts)
        assert len(results) == 3
        for r in results:
            assert "label" in r

    def test_empty_batch(self):
        from naijaml.nlp import analyze_sentiment_batch
        assert analyze_sentiment_batch([]) == []

    def test_single_item_batch(self):
        from naijaml.nlp import analyze_sentiment_batch
        results = analyze_sentiment_batch(["hello"])
        assert len(results) == 1


class TestIsAvailable:
    """Test is_available check."""

    def test_returns_bool(self):
        from naijaml.nlp import is_sentiment_available
        result = is_sentiment_available()
        assert isinstance(result, bool)



"""NaijaML NLP utilities for Nigerian languages."""
from naijaml.nlp.preprocess import (
    # Unicode
    normalize_unicode,
    strip_diacritics,
    # Social media
    clean_social_media,
    extract_hashtags,
    extract_mentions,
    # PII
    mask_pii,
    find_phones,
    find_naira_amounts,
    # Nigerian-specific
    normalize_naira_symbol,
    clean_nigerian_text,
    # Pidgin handling
    is_pidgin_particle,
    preserve_pidgin_particles,
    get_pidgin_particles,
)
from naijaml.nlp.langdetect import (
    detect_language,
    detect_language_with_confidence,
    detect_all_languages,
    SUPPORTED_LANGUAGES,
    train_and_save_model,
    train_and_save_profiles,  # backwards compatibility alias
)
from naijaml.nlp.diacritizer import (
    diacritize,
    diacritize as diacritize_yoruba,  # explicit alias
    diacritize_dot_below,
    diacritize_dot_below as diacritize_yoruba_dot_below,  # explicit alias
    syllabify,
    syllabify as syllabify_yoruba,  # explicit alias
    train_and_save_model as train_diacritizer,
    train_and_save_model as train_yoruba_diacritizer,  # explicit alias
    train_and_save_dot_below_model,
    evaluate_dot_below_accuracy,
    evaluate_full_diacritization_accuracy,
    compare_diacritization_methods,
    strip_diacritics as strip_yoruba_diacritics,  # expose tones_only option
)
from naijaml.nlp.igbo_diacritizer import (
    diacritize_igbo,
    syllabify as syllabify_igbo,
    train_and_save_model as train_igbo_diacritizer,
)
from naijaml.nlp.evaluation import (
    evaluate_diacritizer,
    DiacritizerMetrics,
    run_all_evaluations,
)
from naijaml.nlp.sentiment import (
    analyze_sentiment,
    get_sentiment,
    get_sentiment_with_confidence,
    analyze_batch as analyze_sentiment_batch,
    is_available as is_sentiment_available,
)
from naijaml.nlp.tokenizer import (
    Tokenizer,
    tokenize,
    count_tokens,
    get_supported_languages as get_tokenizer_languages,
    SUPPORTED_LANGUAGES as TOKENIZER_LANGUAGES,
)

__all__ = [
    # Unicode
    "normalize_unicode",
    "strip_diacritics",
    # Social media
    "clean_social_media",
    "extract_hashtags",
    "extract_mentions",
    # PII
    "mask_pii",
    "find_phones",
    "find_naira_amounts",
    # Nigerian-specific
    "normalize_naira_symbol",
    "clean_nigerian_text",
    # Pidgin handling
    "is_pidgin_particle",
    "preserve_pidgin_particles",
    "get_pidgin_particles",
    # Language detection
    "detect_language",
    "detect_language_with_confidence",
    "detect_all_languages",
    "SUPPORTED_LANGUAGES",
    "train_and_save_model",
    "train_and_save_profiles",  # backwards compatibility alias
    # Yorùbá diacritization
    "diacritize",
    "diacritize_yoruba",
    "diacritize_dot_below",
    "diacritize_yoruba_dot_below",
    "syllabify",
    "syllabify_yoruba",
    "train_diacritizer",
    "train_yoruba_diacritizer",
    "train_and_save_dot_below_model",
    "strip_yoruba_diacritics",
    # Diacritizer evaluation
    "evaluate_dot_below_accuracy",
    "evaluate_full_diacritization_accuracy",
    "compare_diacritization_methods",
    # Igbo diacritization
    "diacritize_igbo",
    "syllabify_igbo",
    "train_igbo_diacritizer",
    # Evaluation
    "evaluate_diacritizer",
    "DiacritizerMetrics",
    "run_all_evaluations",
    # Sentiment analysis
    "analyze_sentiment",
    "get_sentiment",
    "get_sentiment_with_confidence",
    "analyze_sentiment_batch",
    "is_sentiment_available",
    # Tokenizer
    "Tokenizer",
    "tokenize",
    "count_tokens",
    "get_tokenizer_languages",
    "TOKENIZER_LANGUAGES",
]



"""Tests for Nigerian text preprocessing module."""
from __future__ import annotations

import pytest


class TestNormalizeUnicode:
    """Test Unicode normalization."""

    def test_nfc_default(self):
        from naijaml.nlp import normalize_unicode
        # NFC should compose characters
        result = normalize_unicode("ọjọ́")
        assert isinstance(result, str)

    def test_empty_string(self):
        from naijaml.nlp import normalize_unicode
        assert normalize_unicode("") == ""

    def test_ascii_unchanged(self):
        from naijaml.nlp import normalize_unicode
        assert normalize_unicode("hello world") == "hello world"


class TestStripDiacritics:
    """Test diacritic stripping."""

    def test_yoruba_diacritics(self):
        from naijaml.nlp import strip_diacritics
        result = strip_diacritics("Ojó lo sí ọjà lánà")
        assert "ọ" not in result
        assert "á" not in result
        assert "ó" not in result

    def test_plain_text_unchanged(self):
        from naijaml.nlp import strip_diacritics
        assert strip_diacritics("hello world") == "hello world"

    def test_empty(self):
        from naijaml.nlp import strip_diacritics
        assert strip_diacritics("") == ""


class TestCleanSocialMedia:
    """Test social media cleaning."""

    def test_removes_urls(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("Check https://t.co/abc")
        assert "https" not in result

    def test_removes_mentions(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("@user hello")
        assert "@user" not in result
        assert "hello" in result

    def test_keeps_hashtags_by_default(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("Great #Nollywood film")
        assert "#Nollywood" in result

    def test_removes_hashtags_when_asked(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("Great #Nollywood film", remove_hashtags=True)
        assert "#Nollywood" not in result

    def test_reduces_repeated_chars(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("Greaaaaat!")
        assert "aaaa" not in result

    def test_lowercase(self):
        from naijaml.nlp import clean_social_media
        result = clean_social_media("HELLO World", lowercase=True)
        assert result == "hello world"


class TestExtractHashtags:
    """Test hashtag extraction."""

    def test_extracts_hashtags(self):
        from naijaml.nlp import extract_hashtags
        result = extract_hashtags("Great #Nollywood #film")
        assert "Nollywood" in result
        assert "film" in result

    def test_no_hashtags(self):
        from naijaml.nlp import extract_hashtags
        assert extract_hashtags("No hashtags here") == []


class TestExtractMentions:
    """Test mention extraction."""

    def test_extracts_mentions(self):
        from naijaml.nlp import extract_mentions
        result = extract_mentions("cc @user1 @user2")
        assert "user1" in result
        assert "user2" in result

    def test_no_mentions(self):
        from naijaml.nlp import extract_mentions
        assert extract_mentions("No mentions here") == []


class TestMaskPII:
    """Test PII masking."""

    def test_masks_phone(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("Call 08012345678")
        assert "[PHONE]" in result
        assert "08012345678" not in result

    def test_masks_email(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("Email user@example.com")
        assert "[EMAIL]" in result
        assert "user@example.com" not in result

    def test_masks_bvn(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("BVN: 22123456789")
        assert "[BVN]" in result

    def test_no_pii_unchanged(self):
        from naijaml.nlp import mask_pii
        text = "Lagos is a city"
        assert mask_pii(text) == text

    def test_custom_masks(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("Call 08012345678", phone_mask="***")
        assert "***" in result

    def test_naira_not_masked_by_default(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("Price is ₦5,000")
        assert "₦5,000" in result

    def test_naira_masked_when_enabled(self):
        from naijaml.nlp import mask_pii
        result = mask_pii("Price is ₦5,000", mask_naira=True)
        assert "[AMOUNT]" in result


class TestFindPhones:
    """Test phone number finder."""

    def test_finds_standard_number(self):
        from naijaml.nlp import find_phones
        result = find_phones("Call 08012345678")
        assert len(result) >= 1

    def test_no_phones(self):
        from naijaml.nlp import find_phones
        assert find_phones("No phone here") == []


class TestFindNairaAmounts:
    """Test Naira amount finder."""

    def test_finds_naira_symbol(self):
        from naijaml.nlp import find_naira_amounts
        result = find_naira_amounts("Price is ₦5,000")
        assert len(result) >= 1

    def test_no_amounts(self):
        from naijaml.nlp import find_naira_amounts
        assert find_naira_amounts("No money here") == []


class TestNormalizeNairaSymbol:
    """Test Naira symbol normalization."""

    def test_ngn_to_naira(self):
        from naijaml.nlp import normalize_naira_symbol
        result = normalize_naira_symbol("NGN5000")
        assert "₦" in result

    def test_already_normalized(self):
        from naijaml.nlp import normalize_naira_symbol
        assert "₦" in normalize_naira_symbol("₦5000")


class TestCleanNigerianText:
    """Test all-in-one cleaning function."""

    def test_full_pipeline(self):
        from naijaml.nlp import clean_nigerian_text
        text = "@user Check https://t.co/abc Ọjọ́ is great!"
        result = clean_nigerian_text(text)
        assert "@user" not in result
        assert "https" not in result
        assert "Ọjọ́" in result

    def test_with_pii_masking(self):
        from naijaml.nlp import clean_nigerian_text
        text = "Call 08012345678"
        result = clean_nigerian_text(text, mask_pii_data=True)
        assert "[PHONE]" in result


class TestPidginHandling:
    """Test Pidgin particle handling."""

    def test_is_pidgin_particle(self):
        from naijaml.nlp import is_pidgin_particle
        assert is_pidgin_particle("sha") is True
        assert is_pidgin_particle("abeg") is True
        assert is_pidgin_particle("computer") is False

    def test_preserve_pidgin_particles(self):
        from naijaml.nlp import preserve_pidgin_particles
        result = preserve_pidgin_particles("The film sha too sweet", ["the", "sha", "too"])
        assert "sha" in result

    def test_get_pidgin_particles(self):
        from naijaml.nlp import get_pidgin_particles
        particles = get_pidgin_particles()
        assert isinstance(particles, set)
        assert "abeg" in particles



"""Evaluate tokenizers: fertility rate, coverage, roundtrip fidelity, diacritic preservation.

Results saved to evaluation_reports/.
"""
from __future__ import annotations

import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger(__name__)

PROJECT_ROOT = Path(__file__).parent.parent
FIXTURES_DIR = PROJECT_ROOT / "tests" / "fixtures"
REPORTS_DIR = PROJECT_ROOT / "evaluation_reports"
REPORTS_DIR.mkdir(exist_ok=True)


def load_test_texts() -> Dict[str, List[str]]:
    """Load test texts per language from fixtures."""
    texts = {}

    # Language samples
    lang_file = FIXTURES_DIR / "language_samples.json"
    if lang_file.exists():
        with open(lang_file) as f:
            texts = json.load(f)

    # Add diacritized Yoruba
    yor_file = FIXTURES_DIR / "eval" / "unseen_yoruba_sentences.json"
    if yor_file.exists():
        with open(yor_file) as f:
            data = json.load(f)
        texts.setdefault("yor", []).extend(data["sentences"][:20])

    # Add diacritized Igbo
    ibo_file = FIXTURES_DIR / "eval" / "unseen_igbo_sentences.json"
    if ibo_file.exists():
        with open(ibo_file) as f:
            data = json.load(f)
        texts.setdefault("ibo", []).extend(data["sentences"][:20])

    return texts


def evaluate_tokenizer_for_lang(lang: str, texts: List[str]) -> Dict:
    """Evaluate tokenizer for a specific language."""
    from naijaml.nlp.tokenizer import Tokenizer

    # Map fixture lang codes to tokenizer lang codes
    lang_map = {"yor": "yoruba", "hau": "hausa", "ibo": "igbo", "pcm": "pidgin", "eng": "naija"}
    tok_lang = lang_map.get(lang)
    if not tok_lang:
        return {"error": f"No tokenizer for {lang}"}

    try:
        tok = Tokenizer(tok_lang)
    except Exception as e:
        return {"error": str(e)}

    total_words = 0
    total_tokens = 0
    roundtrip_pass = 0
    roundtrip_fail = 0
    diacritic_preserved = 0
    diacritic_total = 0
    roundtrip_errors = []

    diacritic_chars = set("ọẹṣáàéèíìóòúùọ́ọ̀ẹ́ẹ̀ịụñ")

    for text in texts:
        words = text.split()
        total_words += len(words)

        # Encode
        token_ids = tok.encode(text)
        total_tokens += len(token_ids)

        # Roundtrip fidelity
        decoded = tok.decode(token_ids)
        if decoded.strip() == text.strip():
            roundtrip_pass += 1
        else:
            roundtrip_fail += 1
            if len(roundtrip_errors) < 10:
                roundtrip_errors.append({
                    "original": text[:80],
                    "decoded": decoded[:80],
                })

        # Diacritic preservation check
        for char in text:
            if char.lower() in diacritic_chars:
                diacritic_total += 1
                if char in decoded:
                    diacritic_preserved += 1

    fertility = total_tokens / total_words if total_words > 0 else 0
    roundtrip_rate = roundtrip_pass / (roundtrip_pass + roundtrip_fail) if (roundtrip_pass + roundtrip_fail) > 0 else 0
    diacritic_rate = diacritic_preserved / diacritic_total if diacritic_total > 0 else 1.0

    return {
        "language": lang,
        "total_texts": len(texts),
        "total_words": total_words,
        "total_tokens": total_tokens,
        "fertility_rate": round(fertility, 4),
        "roundtrip_fidelity": round(roundtrip_rate, 4),
        "roundtrip_pass": roundtrip_pass,
        "roundtrip_fail": roundtrip_fail,
        "diacritic_preservation": round(diacritic_rate, 4),
        "diacritic_chars_tested": diacritic_total,
        "roundtrip_errors": roundtrip_errors,
    }


def evaluate_tokenizers() -> Dict:
    """Run tokenizer evaluation across all languages."""
    texts_by_lang = load_test_texts()
    if not texts_by_lang:
        logger.error("No test texts found!")
        return {}

    results = {}
    for lang, texts in texts_by_lang.items():
        logger.info("Evaluating tokenizer for %s (%d texts)", lang, len(texts))
        results[lang] = evaluate_tokenizer_for_lang(lang, texts)

    return {
        "module": "tokenizer",
        "date": datetime.now().isoformat(),
        "per_language": results,
    }


def print_report(result: Dict) -> None:
    """Print human-readable report."""
    if not result:
        return

    print("\n" + "=" * 60)
    print("TOKENIZER EVALUATION")
    print("=" * 60)

    print(f"\n{'Language':<10} {'Fertility':<12} {'Roundtrip':<12} {'Diacritics':<12} {'Texts':<8}")
    print("-" * 54)

    for lang, metrics in result.get("per_language", {}).items():
        if "error" in metrics:
            print(f"{lang:<10} ERROR: {metrics['error']}")
            continue
        print(f"{lang:<10} {metrics['fertility_rate']:<12.2f} "
              f"{metrics['roundtrip_fidelity']:<12.1%} "
              f"{metrics['diacritic_preservation']:<12.1%} "
              f"{metrics['total_texts']:<8}")

    # Show roundtrip errors if any
    for lang, metrics in result.get("per_language", {}).items():
        errors = metrics.get("roundtrip_errors", [])
        if errors:
            print(f"\nRoundtrip errors for {lang}:")
            for err in errors[:3]:
                print(f"  Original: {err['original'][:60]}")
                print(f"  Decoded:  {err['decoded'][:60]}")
                print()


def main():
    result = evaluate_tokenizers()
    print_report(result)

    report_path = REPORTS_DIR / f"{datetime.now().strftime('%Y%m%d')}_tokenizer.json"
    with open(report_path, "w") as f:
        json.dump(result, f, indent=2, ensure_ascii=False)
    logger.info("Report saved to %s", report_path)

    return result


if __name__ == "__main__":
    main()
